#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Service and command classes used for scheduling.

    - AuraScheduler: Threaded service which performs the scheduling.
    - TimeslotCommand: Inherits `EngineExecuter` to perform timeslot commands.
    - PlayCommand: Inherits `EngineExecuter` to perform play commands on playlist items.
"""

from __future__ import annotations

import logging
import threading
import time

import confuse

import aura_engine.engine as engine
import aura_engine.scheduling.timetable as timetable
import aura_engine.scheduling.utils as utils
from aura_engine.base.config import AuraConfig
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.control import EngineExecutor
from aura_engine.core.channels import LoadSourceException
from aura_engine.resources import ResourceClass, ResourceUtil
from aura_engine.scheduling.domain import PlaylistItem, PlayState, Timeslot

#
#   EngineExecutor Commands
#


class TimeslotCommand(EngineExecutor):
    """
    Command for triggering start and end of timeslot events.
    """

    engine: engine.Engine
    config: confuse.Configuration

    def __init__(self, engine: engine.Engine, timeslot: Timeslot):
        """
        Initialize the timeslot command.

        Args:
            engine (engine.Engine): The engine
            timeslot (Timeslot): The timeslot which is starting at this time
        """
        self.config = AuraConfig.instance.config
        self.engine = engine

        now = SU.timestamp()
        fade_out_time = float(self.config.scheduler.fade_out_time)
        start_fade_in = timeslot.get_start() - now
        start_fade_out = timeslot.get_end() - now - fade_out_time
        self.logger.debug(
            f"Fading in timeslot in {start_fade_in} seconds at {SU.fmt_time(timeslot.get_start())}"
            f" | Timeslot: {timeslot}"
        )
        self.logger.debug(
            f"Fading out timeslot in {start_fade_out} seconds at"
            f" {SU.fmt_time(timeslot.get_end() - fade_out_time)} | Timeslot: {timeslot}"
        )
        # Initialize the "fade in" EngineExecuter and instantiate a connected child EngineExecuter
        # for "fade out" when the parent is ready
        id = timeslot.get_id()
        super().__init__(
            f"TIMESLOT#{id}",
            None,
            timeslot.get_start(),
            self.do_start_timeslot,
            timeslot,
        )
        EngineExecutor(
            f"TIMESLOT#{id}",
            self,
            timeslot.get_end() - fade_out_time,
            self.do_end_timeslot,
            timeslot,
        )

    def do_start_timeslot(self, timeslot: Timeslot):
        """
        Indicate the start of the timeslot by sending a `on_timeslot_start` event.
        """
        self.logger.info(SU.cyan(f"=== on_timeslot_start('{timeslot}') ==="))
        self.engine.event_dispatcher.on_timeslot_start(timeslot)

    def do_end_timeslot(self, timeslot: Timeslot):
        """
        Indicate the start of the timeslot by sending a `on_timeslot_end` event.

        Also resetting the used channel.
        """
        self.logger.info(SU.cyan(f"=== on_timeslot_end('{timeslot}') ==="))
        self.engine.event_dispatcher.on_timeslot_end(timeslot)

        def has_direct_successor():
            timetable: timetable.TimetableService = self.engine.scheduler.timetable
            next_timeslot: Timeslot = timetable.get_next_timeslots(1)
            if next_timeslot:
                next_timeslot = next_timeslot[0]
                if next_timeslot.get_start() > timeslot.get_end():
                    return False
                return True
            else:
                return False

        if not has_direct_successor():
            self.engine.player.stop(engine.Player.TransitionType.FADE)


class PlayCommand(EngineExecutor):
    """
    Command for triggering timed preloading and playing as a child command.
    """

    engine: engine.Engine
    config: confuse.Configuration

    def __init__(self, engine: engine.Engine, items: list[PlaylistItem]):
        """Initialize the play command.

        Args:
            engine (engine.Engine): The engine
            items ([PlaylistItem]): One or more playlist items to be started
        """
        self.config = AuraConfig.instance.config
        self.engine = engine

        preload_offset = self.config.scheduler.preload_offset
        first_item: PlaylistItem = items[0]
        start_preload = first_item.get_start() - preload_offset
        # TODO: Question: could get_start() return int instead of float?
        start_play = int(first_item.get_start())
        msg = (
            f"Preloading items at {SU.fmt_time(start_preload)}, {preload_offset} seconds before"
            f" playing it at {SU.fmt_time(start_play)}"
        )
        self.logger.info(msg)
        # Initialize the "preload" EngineExecuter and attach a child `PlayCommand` to the
        # "on_ready" event handler
        id = first_item.get_playlist().get_timeslot().get_id()
        super().__init__(f"PRELOAD#{id}", None, start_preload, self.do_preload, items)
        EngineExecutor(f"PLAY#{id}", self, start_play, self.do_play, items)

    def do_preload(self, items: list[PlaylistItem]):
        """
        Preload the items.

        Args:
            items ([PlaylistItem]): The set of playlist items to be pre-loaded.
        """
        items_str = ResourceUtil.get_items_string(items)
        try:
            first_item: PlaylistItem = items[0]
            if first_item.get_content_type() in ResourceClass.FILE.types:
                self.logger.info(SU.cyan(f"=== preload_group('{items_str}') ==="))
                self.engine.player.preload_group(items)
            else:
                self.logger.info(SU.cyan(f"=== preload('{items_str}') ==="))
                self.engine.player.preload(items[0])
        except LoadSourceException:
            self.logger.critical(SU.red(f"Could not preload items {items_str}"))

        last_item: PlaylistItem = items[-1]
        if not last_item.play.is_ready():
            msg = f"Items didn't reach 'ready' state during preloading (Items: {items_str})"
            last_item.play.state = PlayState.PlayStateType.TIMEOUT
            self.logger.warning(SU.red(msg))

    def do_play(self, items: list[PlaylistItem]):
        """
        Play the items.

        Args:
            items ([PlaylistItem]): The set of playlist items to be played.
        """
        log_interval = 3
        last_log = 0

        items_str = ResourceUtil.get_items_string(items)
        self.logger.info(SU.cyan(f"=== play('{items_str}') ==="))

        last_item: PlaylistItem = items[-1]
        # NOTE: this may not break and keep on displaying messages if the play state
        # is never set to READY or TIMEOUT.
        while not last_item.play.is_ready():
            now = SU.timestamp()
            if now - last_log > log_interval:
                self.logger.critical(SU.red(f"PLAY: Item(s) not yet ready ({items_str})"))
                last_log = now
            if last_item.play.state == PlayState.PlayStateType.TIMEOUT:
                self.logger.warning(SU.red("PLAY: Preloading timed out."))
                break
            time.sleep(1)
        self.engine.player.play(items[0], engine.Player.TransitionType.FADE)
        timetable_renderer: utils.TimetableRenderer = self.engine.scheduler.timetable_renderer
        self.logger.info(timetable_renderer.get_ascii_timeslots())


#
#   Scheduler
#


class AuraScheduler(threading.Thread):
    """
    The Scheduler.

    The program scheduler has two main duties:
        - Retrieve scheduling data from the API, stored in the timetable.
        - Execute engine action commands in an automated, timed fashion.
    """

    # Command class initialization. Can be overridden in test cases.
    TimeslotCommandClass: EngineExecutor = TimeslotCommand
    PlayCommandClass: EngineExecutor = PlayCommand

    config: confuse.Configuration
    logger: logging.Logger
    engine: engine.Engine
    timetable: timetable.TimetableService
    timetable_renderer: utils.TimetableRenderer
    exit_event: threading.Event
    is_initialized: bool
    is_engine_ready: bool

    def __init__(self, engine: engine.Engine):
        """
        Initialize the scheduler.

        Args:
            engine (engine.Engine): The engine.
        """
        self.config = AuraConfig.instance.config
        self.logger = logging.getLogger("engine")
        self.timetable = timetable.TimetableService(self.config.general.cache_dir)
        self.timetable_renderer = utils.TimetableRenderer(self)
        self.engine = engine
        self.engine.scheduler = self
        # FIXME: is_soundsytem_init is never used
        self.is_soundsytem_init = False
        self.exit_event: threading.Event

        # Scheduler Initialization
        self.is_initialized = False
        self.is_engine_ready = False

    def boot(self):
        """
        Start the scheduler in a thread.
        """
        # Init scheduling thread
        threading.Thread.__init__(self)
        self.exit_event = threading.Event()
        self.start()

    def run(self):
        """
        Execute the scheduler.

        Called when thread is started via `start()`. It does the following:

            1. Refresh timetable periodically from the API depending on the
                `fetching_frequency` defined in the engine configuration.
            2. Merge the timetable with the latest, local version.
            3. Queue all timeslots of the timetable

        On every cycle the configuration file is reloaded, to allow modifications while running
        the engine.
        """
        while not self.exit_event.is_set():
            try:
                AuraConfig.instance.load_config()
                seconds_to_wait = self.config.scheduler.fetching_frequency
                msg = f"== start fetching new timeslots (every {seconds_to_wait} seconds) =="
                self.logger.info(SU.cyan(msg))

                # Load some stuff from the API in any case
                self.timetable.refresh()

                # Queue only when the engine is ready to play
                if self.is_initialized:
                    self.queue_program()

            except Exception as e:
                msg = f"Unhandled error while fetching & scheduling new program! ({str(e)})"
                self.logger.critical(SU.red(msg), e)
                # Keep on working anyway

            EngineExecutor.log_commands()
            self.exit_event.wait(seconds_to_wait)

    #
    #   EVENT HANDLERS
    #

    def on_ready(self):
        """
        Handle `on_ready` event.

        Called when the engine has finished booting and is ready to play.

            1. First the item currently to be played is scheduled.
            2. All following items are queued.
        """
        self.is_initialized = True
        self.logger.info(self.timetable_renderer.get_ascii_timeslots())

        self.play_active_item()
        self.queue_startup_items()

    #
    #   METHODS
    #

    def get_timetable(self) -> timetable.TimetableService:
        """
        Get the timetable service.

        Returns:
            (TimetableService): The timetable service.
        """
        return self.timetable

    def play_active_item(self):
        """
        Play currently active playlist item, as per timetable.

        Plays the item scheduled for the very current moment and forwards to the scheduled
        position in time. Usually called when the Engine boots.
        """
        sleep_offset = 10
        active_timeslot: Timeslot = self.timetable.get_current_timeslot()

        # Schedule any available fallback playlist
        if active_timeslot:
            if active_timeslot.is_virtual():
                self.logger.info(f"Skipping virtual timeslot: {active_timeslot}")
            else:
                # Create command timer to indicate the start of the timeslot
                AuraScheduler.TimeslotCommandClass(self.engine, active_timeslot)

        active_item: PlaylistItem | None = self.timetable.get_current_item()
        if not active_item:
            return

        # In case of a file-system source, we need to fast-forward to the current marker as per
        # position in the timeslot
        if active_item.get_content_type() in ResourceClass.FILE.types:
            # Calculate the seconds we have to fast-forward
            now = engine.Engine.engine_time()
            seconds_to_roll = now - active_item.get_start()

            # If the roll exceeds the length of the current track,
            # there's no need to do anything - the scheduler takes care of the rest
            if (seconds_to_roll + sleep_offset) > active_item.duration:
                msg = "The FFWD [>>] range exceeds the length of the item. \
                    Drink some tea and wait for the sound of the next item."
                self.logger.info(msg)
                return
            else:
                # Preload and play active item
                AuraScheduler.PlayCommandClass(self.engine, [active_item])

                # Fast-forward to the scheduled position
                # TODO The roll should happen before the channel is active, avoiding the roll being
                # audible.
                if seconds_to_roll > 0:
                    # Without plenty of timeout (10s) the roll doesn't work
                    # TODO Check if this is still the case with Liquidsoap 2
                    # TODO Think if this should be externalized as a command for better testability
                    def async_preroll(seconds_to_roll):
                        seconds_to_roll += sleep_offset
                        time.sleep(sleep_offset)
                        self.logger.info(f"Going to fast-forward {seconds_to_roll} seconds")
                        rolled = self.engine.player.roll(active_item.play.channel, seconds_to_roll)
                        if rolled:
                            self.logger.info("Pre-roll done")

                    thread = threading.Thread(target=async_preroll, args=(seconds_to_roll,))
                    thread.start()

        elif (
            active_item.get_content_type() in ResourceClass.STREAM.types
            or active_item.get_content_type() in ResourceClass.LIVE.types
        ):
            # Preload and play active item
            AuraScheduler.PlayCommandClass(self.engine, [active_item])

        else:
            self.logger.critical(f"Unknown content type for item {active_item}")
            return

    def queue_program(self):
        """
        Queue the current program.

        Playlists of every timeslot are queued by creating timed commands to the sound-system to
        enable the individual tracks of playlists.
        """
        # Get a clean set of the timeslots within the scheduling window
        timeslots = self.timetable.get_next_timeslots(window_aware=True)

        # Queue the timeslots, their playlists and items
        if not timeslots:
            self.logger.info(SU.green("No program to be queued."))
        else:
            next_timeslot: Timeslot
            for next_timeslot in timeslots:
                if next_timeslot.is_virtual():
                    self.logger.info(f"Skipping virtual timeslot: {next_timeslot}")
                    continue
                # Create command timer to indicate the start of the timeslot
                AuraScheduler.TimeslotCommandClass(self.engine, next_timeslot)

                playlist = next_timeslot.get_current_playlist()
                if playlist:
                    self.queue_playlist_items(next_timeslot, playlist.items)
            self.logger.info(SU.green("Finished queuing program."))

    def queue_startup_items(self):
        """
        Queue all items after the one currently playing upon startup.

        Don't use this method in any other scenario, as it doesn't respect the scheduling window.
        """
        current_timeslot: Timeslot = self.timetable.get_current_timeslot()

        # Queue the (rest of the) currently playing timeslot upon startup
        if current_timeslot:
            current_playlist = current_timeslot.get_current_playlist()

            if current_playlist:
                active_item: PlaylistItem = self.timetable.get_current_item()
                if active_item:
                    # Queue open items for current playlist
                    rest_of_playlist = active_item.get_all_next(True)
                    self.queue_playlist_items(current_timeslot, rest_of_playlist)

    def queue_playlist_items(self, timeslot: Timeslot, items: list[PlaylistItem]):
        """
        Create player commands for all playlist items to be executed at the scheduled time.

        Since each scheduled playlist can consist of multiple item types such as *file*, *live*,
        and *stream*, the play-out of the timeslot is actually a bit more complex.
        Before any playlist items of the timeslot can be turned into sound, they need to be
        aggregated, queued and pre-loaded.

        1. First, all items are aggregated when they hold filesystem items.
            Given you have a playlist with 10 items, the first 4 are consisting of files, the
            next two of a a stream and a live source. The last 4 are files again.
            These items are now aggregated into 4 groups: one for the files, one for the stream,
            one for the live item and another one for files.
            For each group a timer for executing the next step is created.

        2. Now, the playlist items are going to be "pre-loaded".
            This means that filesystem items are queued and pre-loaded and items which are
            based on audio streams are buffered.
            This is required to allow a seamless play-out, when its time to do so (in the next
            stage).
            Due to their nature, playlist items which hold live audio sources are not affected by
            this stage at all.

        Args:
            timeslot (Timeslot): The timeslot this items belong to.
            items ([PlaylistItem]): The playlist items to be scheduled for playout.
        """
        item_groups = []
        item_groups.append([])
        index = 0

        # Group/aggregate all filesystem items, allowing them to be queued at once
        item: PlaylistItem
        previous_item: PlaylistItem = None
        for item in items:
            if previous_item is None or (
                previous_item.get_content_type() == item.get_content_type()
                and item.get_content_type() in ResourceClass.FILE.types
            ):
                item_groups[index].append(item)
            else:
                index += 1
                item_groups.append([])
                item_groups[index].append(item)
            previous_item = item
        self.logger.info(f"Built {len(item_groups)} item group(s)")

        # Timeslot function calls
        if len(items) > 0 and len(item_groups) > 0:
            for items in item_groups:
                if not isinstance(items, list):
                    raise ValueError(f"Invalid Item Group: {str(items)}")

                # Create command timers for each item group
                AuraScheduler.PlayCommandClass(self.engine, items)
        else:
            self.logger.warn(SU.red(f"Nothing to schedule for timeslot {timeslot}"))

    def terminate(self):
        """
        Call when thread is stopped or termination signal is received.
        """
        self.logger.info(SU.yellow("[Scheduler] Shutting down..."))
        self.timetable.terminate()
        self.exit_event.set()
